// Copyright (C) 2022 - 2023, Xilinx Inc - All rights reserved
//
// Licensed under the Apache License, Version 2.0 (the "License"). You may
// not use this file except in compliance with the License. A copy of the
// License is located at
//
//      http://www.apache.org/licenses/LICENSE-2.0
//
// Unless required by applicable law or agreed to in writing, software
// distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
// WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
// License for the specific language governing permissions and limitations
// under the License.

#include <cstring>
#include <assert.h>
#include "create_json.h"
#include "xrm_utils.h"
#include "log_ama.h"

XrmUtilsIPMapping xrm_ip_mapping[] = {{"SCALER", "xrmMA35ScalePlugin", "scaler", "SCALER_AMA"}, {"DECODER", "xrmMA35DecodePlugin", "decoder", "DECODER_AMA"},
    {"ENCODER", "xrmMA35EncodePlugin", "encoder", "ENCODER_TYPE1_AMA"}, {"ENCODER", "xrmMA35EncodePlugin", "encoder", "ENCODER_TYPE2_AMA"},
    {"LOOKAHEAD", "xrmMA35EncodePlugin", "lookahead", "LOOKAHEAD_AMA"}, {"ML", "xrmMA35MLPlugin", "ml", "ML_AMA"}};

/**
 * @brief Returns the mapping index for the IP
 *
 * @param ip_name Name of the IP module
 * @return Mapping ID or XRM_ERROR on failure
 */
static int xrm_utils_get_mapping_id(const std::string& ip_name) {
  for (int i = 0; i < (int) (sizeof(xrm_ip_mapping) / sizeof(XrmUtilsIPMapping)); i++) {
    if (xrm_ip_mapping[i].ip_name == ip_name) {
      return i;
    }
  }
  return XRM_ERROR;
}

/**
 * @brief Get the device index using the XRM_RESERVE_ID environment variable.
 * 
 * @return The device index corresponding to XRM_RESERVE_ID, 0 if unset, or
 * XRM_ERROR on error
 */
int xrm_interface_get_dev_index(void) {
  const char* xrm_reserve_id_as_str = getenv("XRM_RESERVE_ID");
  xrmContext  xrm_ctx               = xrmCreateContext(XRM_API_VERSION_1);
  if (!xrm_reserve_id_as_str || strcmp(xrm_reserve_id_as_str, "") == 0) {
    if (xrm_ctx) {
      xrmDestroyContext(xrm_ctx);
    } else {
      LOG_AMA_WARN(LOG_AMA_LAYER_XRM, "Unable to create XRM context. XRM daemon likely not running. Skipping XRM validation");
    }
    return 0;
  }
  if (xrm_ctx == NULL) {
    LOG_AMA_ERROR(LOG_AMA_LAYER_XRM, "Unable to create XRM context. XRM daemon likely not running, but XRM_RESERVE_ID is set!");
    return XRM_ERROR;
  }

  char* endptr;
  errno              = 0;
  int xrm_reserve_id = strtol(xrm_reserve_id_as_str, &endptr, 0);
  //check for strtol errors
  if (errno != 0 || xrm_reserve_id < 0) {
    perror("strtol");
    LOG_AMA_ERROR(LOG_AMA_LAYER_XRM, "Failed to parse XRM_RESERVE_ID");
    xrmDestroyContext(xrm_ctx);
    return XRM_ERROR;
  }

  xrmCuPoolResourceV2 query_transcode_cu_pool_res;
  memset(&query_transcode_cu_pool_res, 0, sizeof(query_transcode_cu_pool_res));

  xrmReservationQueryInfoV2 reserveQueryInfo;
  memset(&reserveQueryInfo, 0, sizeof(xrmReservationQueryInfoV2));
  reserveQueryInfo.poolId = xrm_reserve_id;

  if (xrmReservationQueryV2(xrm_ctx, &reserveQueryInfo, &query_transcode_cu_pool_res) != 0) {
    LOG_AMA_ERROR(LOG_AMA_LAYER_XRM, "Failed to query allocated cu list");
    xrmDestroyContext(xrm_ctx);
    return XRM_ERROR;
  }
  xrmDestroyContext(xrm_ctx);
  return query_transcode_cu_pool_res.cuResources[0].deviceId;
}

/**
 * @brief Checks and allocates resources for xrm given the cu pool properties
 * with relevant device info
 * 
 * @param xrm_ctx The xrm context used for xrm calls
 * @param pool_property Pool property containing relevant device / job info
 * @param xrm_cu_list_res Will be filled by xrmCuListAllocV2
 * @return XRM_SUCCESS or XRM_ERROR
 */
int xrm_utils_reserve_resources(xrmContext xrm_ctx, xrmCuPoolPropertyV2* pool_property, xrmCuListResourceV2* xrm_cu_list_res) {
  if (!getenv("XRM_RESERVE_ID")) {
    int num_cu_pool = xrmCheckCuPoolAvailableNumV2(xrm_ctx, pool_property);
    if (num_cu_pool <= 0) {
      LOG_AMA_ERROR(LOG_AMA_LAYER_XRM, "Insufficient resources available for allocation");
      return XRM_ERROR;
    }
    LOG_AMA_INFO(LOG_AMA_LAYER_XRM, "Num CU pools available %d", num_cu_pool);
  }

  if (xrmCuListAllocV2(xrm_ctx, &pool_property->cuListProp, xrm_cu_list_res) != XRM_SUCCESS) {
    LOG_AMA_ERROR(LOG_AMA_LAYER_XRM, "Failed allocating resources from XRM");
    return XRM_ERROR;
  }
  return XRM_SUCCESS;
}

/**
 * @brief Update the cu pool properties based on the relevant information
 * provided
 * 
 * @param ip_name What module this is: "DECODER", "SCALER", "ENCODER" or "ML"
 * @param device_id The device index whose resources we should query
 * @param cu_load The load of the job
 * @param cu_pool_prop The properties to be set
 * @return XRM_SUCCESS or XRM_ERROR
 */
int xrm_utils_update_pool_props(const std::string& ip_name, int device_id, int cu_load, xrmCuPoolPropertyV2* cu_pool_prop, bool is_xav1) {
  int      cu_num   = cu_pool_prop->cuListProp.cuNum;
  uint64_t dev_info = (device_id << XRM_DEVICE_INFO_DEVICE_INDEX_SHIFT) |
      ((uint64_t) XRM_DEVICE_INFO_CONSTRAINT_TYPE_HARDWARE_DEVICE_INDEX << XRM_DEVICE_INFO_CONSTRAINT_TYPE_SHIFT);
  int mapping_id    = xrm_utils_get_mapping_id(ip_name);
  int cu_reserve_id = 0;
  if ((is_xav1 != 1) && (ip_name == "ENCODER")) {
    mapping_id++;
  }
  if (getenv("XRM_RESERVE_ID")) {
    char* endptr;
    errno         = 0;
    cu_reserve_id = (int) strtol(getenv("XRM_RESERVE_ID"), &endptr, 0);
    if (errno != 0) {
      LOG_AMA_ERROR(LOG_AMA_LAYER_XRM, "Failed to parse XRM_RESERVE_ID");
      return XRM_ERROR;
    }
  }

  cu_pool_prop->cuListNum      = 1;
  xrmCuPropertyV2* cu_property = &cu_pool_prop->cuListProp.cuProps[cu_num];
  strcpy(cu_property->kernelName, xrm_ip_mapping[mapping_id].kernel_name.c_str());
  strcpy(cu_property->kernelAlias, xrm_ip_mapping[mapping_id].kernel_alias.c_str());
  cu_property->devExcl                       = false;
  cu_property->requestLoad                   = XRM_UTILS_PRECISION_1000000_BIT_MASK(cu_load);
  cu_property->deviceInfo                    = dev_info;
  cu_property->poolId                        = cu_reserve_id;
  cu_property->logicGroupInfo.constraintInfo = XRM_LOGIC_GROUP_INFO_CONSTRAINT_TYPE_NULL;

  cu_pool_prop->cuListProp.cuNum++;
  return XRM_SUCCESS;
}

// The below function get the requested load and cap it to maximum capacity of a CU, if requested load is more than maximum capacity.
// This is a temporary hack to allow 4kp120, 8kp30 usecases without stopping XRMD service.
// Needs to be handled properly, once XRMD supports Over subscription
static int get_requested_load(const std::string& ip_name, int& cu_load, int cores = 1) {
  cu_load = atoi(static_cast<char*>(strtok(NULL, " ")));
  assert(errno == 0 && "Failed to atoi cu_load");

  // Validate the requested load for 2-core ULL encoding case and exit with error if load exceeds maximum capacity, as this shouldn't be over subscribed
  // Other cases, cap it to maximum capacity and allow the usecase by XRMD. These cases won't meet the realtime performance
  if (cores == 2) {
    if (cu_load > XRM_MAX_CU_LOAD_GRANULARITY_1000000 * cores) {
      LOG_AMA_ERROR(LOG_AMA_LAYER_XRM, "%s requested load(%d) exceeds the maximum capacity(%u)", ip_name.c_str(), cu_load, XRM_MAX_CU_LOAD_GRANULARITY_1000000 * cores);
      return XRM_ERROR;
    }
  } else if (cu_load > XRM_MAX_CU_LOAD_GRANULARITY_1000000) {
    LOG_AMA_WARN(LOG_AMA_LAYER_XRM,
        "Slower than realtime due to %s requested load(%d) exceeds the maximum capacity(%u). Capping it to maximum capacity to allow this usecase by "
        "XRMD",
        ip_name.c_str(), cu_load, XRM_MAX_CU_LOAD_GRANULARITY_1000000);
    // Hack: Cap to CU maximum capacity
    cu_load = XRM_MAX_CU_LOAD_GRANULARITY_1000000;
  }
  return XRM_SUCCESS;
}

/**
 * @brief Calculate the load for the module specified by ip_name. Can be
 * "DECODER", "SCALER", "ENCODER" or "ML"
 * 
 * @param xrm_ctx xrm context used for calls to xrm
 * @param ip_name Can be "DECOER", "SCALER", "ENCODER" or "ML"
 * @param cu_load Pointer to the cu load to be set, in the case of encoder, this
 * is an array of size 3.
 * cu_load[0] = is_xav1 flag
 * cu_load[1] = encoder load
 * cu_load[2] = lookahead load
 * @return The cu_load calculated, or XRM_ERROR on error
 */
int xrm_utils_load_calc(xrmContext xrm_ctx, XrmInterfaceProperties* input_props, XrmInterfaceProperties* output_props, int num_outputs, const std::string& ip_name,
    int* cu_load, bool is_xav1) {
  int         ret;
  int         func_id    = 0;
  int         mapping_id = xrm_utils_get_mapping_id(ip_name);
  std::string plg_param_input;
  create_json(input_props, output_props, num_outputs, ip_name, plg_param_input, is_xav1);
  xrmPluginFuncParam plg_param;
  memset(&plg_param, 0, sizeof(xrmPluginFuncParam));
  strcpy(plg_param.input, plg_param_input.c_str());
  ret = xrmExecPluginFunc(xrm_ctx, const_cast<char*>(xrm_ip_mapping[mapping_id].plugin_name.c_str()), func_id, &plg_param);
  if (ret != XRM_SUCCESS) {
    LOG_AMA_ERROR(LOG_AMA_LAYER_XRM, "XRM plugin failed to calculate load for %s: %d", ip_name.c_str(), ret);
    return XRM_ERROR;
  }

  errno           = 0;
  int num_devices = atoi(static_cast<char*>(strtok(plg_param.output, " ")));
  assert(errno == 0 && "Failed to atoi num devices");
  assert(num_devices == 1 && "Number of devices for job is not 1. This shouldn't happen in this code.\n");
  (void) num_devices;

  int num_channels = atoi(static_cast<char*>(strtok(NULL, " ")));
  assert(errno == 0 && "Failed to atoi num channels");
  assert(num_channels == 1 && "xrm load calc found num channels is not one. This shouldn't happen in this code.\n");
  (void) num_channels;

  cu_load[0] = 0;
  // get Decoder load or Scaler load or ML load or Encoder type flag
  if (get_requested_load(xrm_ip_mapping[mapping_id].ip_name, cu_load[0])) {
    return XRM_ERROR;
  }
  // Return from here, if it is not an ecnoder CU
  if (ip_name != "ENCODER") {
    LOG_AMA_INFO(LOG_AMA_LAYER_XRM, "%s cu_load %d", xrm_ip_mapping[mapping_id].ip_name.c_str(), cu_load[0]);
    return cu_load[0];
  }

  cu_load[1] = 0;
  // For Encoder, cu_load[0] contains is_xav1 flag. if cu_load[0] is 1, type 1 (xav1) encoder, otherwise type 2 (VC8000E) encoder
  // get encoder load
  if (get_requested_load(xrm_ip_mapping[mapping_id].ip_name, cu_load[1], input_props->enc_cores)) {
    return XRM_ERROR;
  }

  cu_load[2] = 0;
  // get lookahead load
  if (get_requested_load("LOOKAHEAD", cu_load[2], input_props->enc_cores)) {
    return XRM_ERROR;
  }

  LOG_AMA_INFO(LOG_AMA_LAYER_XRM, "%s Type %d, cu_load %d, la_load %d", xrm_ip_mapping[mapping_id].ip_name.c_str(), cu_load[0] ? 1 : 2, cu_load[1], cu_load[2]);
  return cu_load[1];
}
